import os
import pickle
from collections import Counter, OrderedDict
from copy import deepcopy



class ModelBase:
    """
    Base class for preparing features
    """

    def __init__(self):
        self.accented_words_file = os.path.join(os.path.dirname(__file__), "master_word_list.pickle")
        self.ordered_accented_words = {}
        self.counter_dicts = []
        self.other_features = []
        self.all_period_features = []
        self.features = OrderedDict()


    def get_accented_words(self):
        with open(self.accented_words_file, 'rb') as f:
            accented_words = pickle.load(f)
            return accented_words


    def create_accented_word_dict(self):
        accented_words = self.get_accented_words()
        unordered_accented_words = Counter(accented_words)
        ordered_accented_words = OrderedDict(unordered_accented_words)
        self.ordered_accented_words = ordered_accented_words


    def create_fresh_word_dict(self):
        ordered_accented_words_copy = deepcopy(self.ordered_accented_words)
        fresh_word_dict = OrderedDict({k:0 for k in ordered_accented_words_copy})
        return fresh_word_dict


    def get_sections_per_period(self):
        counter_dict = self.features["counter_dict"]
        rules_avg = self.features["rules_avg"]
        words_per_line = self.features["words_per_line"]
        avg_syllables_per_line = self.features["avg_syllables_per_line"]
        rule_0 = self.features["rule_0"]
        rule_1 = self.features["rule_1"]
        rule_2 = self.features["rule_2"]
        rule_3 = self.features["rule_3"]
        rule_4 = self.features["rule_4"]
        rule_5 = self.features["rule_5"]
        rule_6 = self.features["rule_6"]
        self.counter_dicts.append(counter_dict)
        self.other_features = [
                                rules_avg, 
                                words_per_line, 
                                avg_syllables_per_line, 
                                rule_0, 
                                rule_1, 
                                rule_2, 
                                rule_3, 
                                rule_4, 
                                rule_5, 
                                rule_6
                                ] 


    def create_accented_word_feature(self):
        for sect in self.counter_dicts:
            sect = sorted([word for word in sect])
            sect_dict = OrderedDict(Counter(sect))
            sect_combined = OrderedDict()
            for k,v in self.ordered_accented_words.items():
                if k in sect_dict:
                    sect_combined[k] = 1
                else:
                    sect_combined[k] = 0
            one_hot_sect = [[float(v) for v in sect_combined.values()]]
            one_hot_sect[0].extend(self.other_features)

        return one_hot_sect


    def combine_all_period_features(self):
        pass


    def main(self):
        pass


import unittest
from ipclassifier.iambic_line_processors import IambicLine
from ipclassifier.dataprep import RawFileProcessor
from ipclassifier.token_processors import Tokenizer
from ipclassifier.utils import DictsSingleton



class TestToIambicLine(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.dicts = DictsSingleton()

    def setUp(self):
        self.ibls = []
        sentences = RawFileProcessor("test/test_data/basic_lines.txt").cleaned_contents
        tokenizer = Tokenizer(sentences, self.dicts)
        line_tokens = tokenizer.create_tokens()
        for line in line_tokens:
            iambic_line = IambicLine(line)
            self.ibls.append(iambic_line)


    def test_simple_case(self):
        is_valid_ip = self.ibls[2]._is_valid_pattern
        self.assertTrue(is_valid_ip) 

    def test_stress_alteration(self):
        is_valid_ip = self.ibls[0]._is_valid_pattern
        self.assertTrue(is_valid_ip)

    def test_stress_promotion(self):
        is_valid_ip = self.ibls[1]._is_valid_pattern
        self.assertTrue(is_valid_ip)

    def test_too_short(self):
        is_valid_ip = self.ibls[3]._is_valid_pattern
        self.assertFalse(is_valid_ip)
    
    def test_demote_stress(self):
        is_valid_ip = self.ibls[4]._is_valid_pattern
        self.assertTrue(is_valid_ip)   

    def test_promote_polysyllabic_stress(self):
        is_valid_ip = self.ibls[5]._is_valid_pattern
        self.assertTrue(is_valid_ip)    

    def test_heaven_optionally_1_syllable(self):
        is_valid_ip = self.ibls[6]._is_valid_pattern
        self.assertTrue(is_valid_ip)    

    def test_non_initial_apostrophe(self):
        is_valid_ip = self.ibls[7]._is_valid_pattern
        self.assertTrue(is_valid_ip)    

    def test_11_syllables(self):
        is_valid_ip = self.ibls[8]._is_valid_pattern
        self.assertTrue(is_valid_ip)  

    def test_compound_false_alarm(self):
        is_valid_ip = self.ibls[9]._is_valid_pattern
        self.assertTrue(is_valid_ip)   

    def test_compound_false_alarm_stress_alteration(self):
        is_valid_ip = self.ibls[10]._is_valid_pattern
        self.assertTrue(is_valid_ip)    

    def test_optional_penultimate_stress_removal(self):
        is_valid_ip = self.ibls[11]._is_valid_pattern
        self.assertTrue(is_valid_ip)  

    def test_optional_penultimate_stress_removal_with_stress_alteration(self):
        is_valid_ip = self.ibls[12]._is_valid_pattern
        self.assertTrue(is_valid_ip)  

    def test_possessive_apostrophe_s_removed(self):
        is_valid_ip = self.ibls[13]._is_valid_pattern
        self.assertTrue(is_valid_ip)    

    def test_anapest(self):
        is_valid_ip = self.ibls[14]._is_valid_pattern
        self.assertFalse(is_valid_ip)    

    def test_multiple_reductions(self):
        is_valid_ip = self.ibls[15]._is_valid_pattern
        self.assertTrue(is_valid_ip)    

    def test_multiple_elided_the(self):
        is_valid_ip = self.ibls[16]._is_valid_pattern
        self.assertTrue(is_valid_ip)    

    def test_ed_ending(self):
        is_valid_ip = self.ibls[17]._is_valid_pattern
        self.assertTrue(is_valid_ip)    


if __name__ == "__main__":
    unittest.main()


import re
from copy import deepcopy
from nltk import pos_tag



class SpellingSyllabifier:
    """
    A last resort if a word token cannot be found elsewhere
    Syllabifies a word based only on its spelling
    Ridiculously expandable
    """

    VOWELS = "aeiouy"
    DUMMY_STRESSED = "AH1"
    DUMMY_UNSTRESSED = "AH0"
    REGEX = {
        # Endings
        "IES": r'ies?$',
        "ION": r'[st]{1}ions?$',
        "ED": r'[^aeiou]ed$',
        "EST": r'.{2,}est$',
        "IZE": r'i[sz]es?$',
        "E": r'[^aeio]es?$',
        "SES": r'[sz]es$',
        "IAN": r'ianS?$',
        # Non-Endings
        "QU": r'qu',
        "AE": r'ae',
        "DOUBLE": r'([eiouy])\1',  # no double a
        "OU": r'ou',
        "EY": r'ey\W?',
        "YV": r'y[aeiou]',
        "EA": r'ea',
        "AI": r'ai',
        "AU": r'au',
        "UI": r'ui',
        "OY": r'oy',
        # Others
        "EOU": r'eou',
        "EON": r'eon',
        "VLV": r'[aeiou]{1}[vwrl][aeiou]{1}'
    }

    def __init__(self, token):
        self.tentative_phonemes = [[]]
        
        self._token = token
        self._syllable_count = 0
        self._modified_word = ''
        self._reduced_syllables = 0

        self._main()

    
    def _get_syllable_count(self):
        word = self._check_endings()
        word = self._check_special_cases(word)
        syllables = [w for w in word if w in self.VOWELS]
        self._syllable_count = len(syllables)


    def _find_multiple(self, regex, word, rev=False):
        res = re.finditer(regex, word)
        indicies = [m.start() + 1 for m in res]
        indicies = indicies[::-1] if rev else indicies
        for idx in indicies:
            word = word[:idx] + word[idx + 1:]
            self._reduced_syllables += 1
        return word


    def _find_single(self, letter, word):
        idx = word.rindex(letter)
        word = word[:idx] + word[idx + 1:]
        self._reduced_syllables += 1
        return word


    def _check_endings(self):
        word = self._token

        if re.search(self.REGEX["EST"], word):
            word = self._find_single("e", word)
            return word

        if re.search(self.REGEX["IZE"], word):
            word = self._find_single("e", word)
            return word

        if re.search(self.REGEX["IAN"], word):
            word = self._find_single("i", word)
            return word

        if re.search(self.REGEX["IES"], word):
            word = self._find_single("e", word)
            return word
       
        if re.search(self.REGEX["SES"], word):
            return word

        if re.search(self.REGEX["E"], word):
            word = self._find_single("e", word)
            return word

        if re.search(self.REGEX["ION"], word)  and len(word) > 4:
            word = self._find_single("i", word)
            return word

        if re.search(self.REGEX["ED"], word)  and len(word) >= 4:
            word = self._find_single("e", word)
            return word
        return word


    def _check_special_cases(self, word=None):
        word = word if word else self._token
    
        if re.search(self.REGEX["AU"], word):
            word = self._find_multiple(self.REGEX["AU"], word)
            return self._check_special_cases(word)

        if re.search(self.REGEX["AI"], word):
            word = self._find_multiple(self.REGEX["AI"], word)
            return self._check_special_cases(word)

        if re.search(self.REGEX["QU"], word):
            word = self._find_multiple(self.REGEX["QU"], word)
            return self._check_special_cases(word)

        if re.search(self.REGEX["AE"], word):
            word = self._find_multiple(self.REGEX["AE"], word, rev=True)
            return self._check_special_cases(word)

        if re.search(self.REGEX["DOUBLE"], word):
            word = self._find_multiple(self.REGEX["DOUBLE"], word)
            return self._check_special_cases(word)

        if re.search(self.REGEX["OU"], word):
            word = self._find_multiple(self.REGEX["OU"], word)
            return self._check_special_cases(word)

        if re.search(self.REGEX["EY"], word):
            word = self._find_multiple(self.REGEX["EY"], word)
            return self._check_special_cases(word)

        if re.search(self.REGEX["YV"], word):
            word = self._find_multiple(self.REGEX["YV"], word)
            return self._check_special_cases(word)

        if re.search(self.REGEX["EA"], word):
            word = self._find_multiple(self.REGEX["EA"], word)
            return self._check_special_cases(word)

        if re.search(self.REGEX["UI"], word):
            word = self._find_multiple(self.REGEX["UI"], word)
            return self._check_special_cases(word)

        if re.search(self.REGEX["OY"], word):
            word = self._find_multiple(self.REGEX["OY"], word)
            return self._check_special_cases(word)
        self._modified_word = word
        return self._modified_word

    
    def _simple_stressor(self, restore_syllables=0):
        count = self._syllable_count
        if count == 1:
            return [[self.DUMMY_STRESSED]]
        else:
            return [[self.DUMMY_STRESSED if i == self._syllable_count - 2 else self.DUMMY_UNSTRESSED for i in range(self._syllable_count + restore_syllables) ]]


    def _complicated_stressor(self, POS, restore_syllables=0):
        if POS == "V" or any([self._token.endswith(ending) for ending in ["est", "eth", "ise", "ize"] ]):
            return [[self.DUMMY_STRESSED if i == 0 else self.DUMMY_UNSTRESSED for i in range(self._syllable_count + restore_syllables) ]]
        if POS in ["N", "J"]:
            return [[self.DUMMY_STRESSED if i == self._syllable_count - 1 else self.DUMMY_UNSTRESSED for i in range(self._syllable_count + restore_syllables) ]]

    
    def _check_ed(self, phonemes):
        if self._token.endswith("ed") and len(self._token) > 4:
            antepenult_letter = self._token[-3]
            if antepenult_letter not in "aeiou":
                phonemes_copy = deepcopy(phonemes[0])
                phonemes_copy.insert(-1,'EH0')
                phonemes.append(phonemes_copy)
                self._reduced_syllables -= 1
        return phonemes

    
    def _check_vowel_cluster(self, phonemes):
        if re.search(self.REGEX["EOU"], self._token):
            phonemes_len = len(phonemes[0])
            reduced_phonemes = [self.DUMMY_STRESSED if i == 0 else self.DUMMY_UNSTRESSED for i in range(phonemes_len - 1) ]
            phonemes.append(reduced_phonemes)
        if re.search(self.REGEX["EON"], self._token):
            reduced_phonemes = [self.DUMMY_STRESSED if i == 0 else self.DUMMY_UNSTRESSED for i in range(2) ]
            phonemes.append(reduced_phonemes)
        if re.search(self.REGEX["VLV"], self._token):
            phonemes_len = len(phonemes[0])
            reduced_phonemes = [self.DUMMY_STRESSED if i == 0 else self.DUMMY_UNSTRESSED for i in range(phonemes_len - 1) ]
            phonemes.append(reduced_phonemes)

        return phonemes


    def _create_phoneme_repr(self):
        """
        Check POS, if none, use simple_stressor, otherwise, use complicated_stressor
        """

        tag = pos_tag([self._token])[0][1]
        if tag.startswith("V") or any([self._token.endswith(ending) for ending in ["est", "eth", "ise", "ize"] ]): #or tag.startswith("N") or tag.startswith("J") 
            self.tentative_phonemes = self._complicated_stressor(tag[0])
            if self._token.endswith('ed'):
                self.tentative_phonemes = self._check_ed(self.tentative_phonemes)
            if self._modified_word:
                self.tentative_phonemes.append(self._complicated_stressor(tag[0], self._reduced_syllables)[0])
        else:
            self.tentative_phonemes = self._simple_stressor()
            if self._token.endswith('ed'):
                self.tentative_phonemes = self._check_ed(self.tentative_phonemes)
            if self._modified_word:
                self.tentative_phonemes.append(self._simple_stressor(self._reduced_syllables)[0])


    def _final_reduction_check(self):
        with_vowel_cluster = self._check_vowel_cluster(self.tentative_phonemes)
        self.tentative_phonemes = with_vowel_cluster


    def _main(self):
        self._get_syllable_count()
        self._create_phoneme_repr()
        self._final_reduction_check()



# slightly modified from https://github.com/kylebgorman/syllabify
from itertools import chain


## constants
SLAX   = {'IH1', 'IH2', 'EH1', 'EH2', 'AE1', 'AE2', 'AH1', 'AH2',
                                                    'UH1', 'UH2',}
VOWELS = {'IY1', 'IY2', 'IY0', 'EY1', 'EY2', 'EY0', 'AA1', 'AA2', 'AA0',
          'ER1', 'ER2', 'ER0', 'AW1', 'AW2', 'AW0', 'AO1', 'AO2', 'AO0',
          'AY1', 'AY2', 'AY0', 'OW1', 'OW2', 'OW0', 'OY1', 'OY2', 'OY0',
          'IH0', 'EH0', 'AE0', 'AH0', 'UH0', 'UW1', 'UW2', 'UW0', 'UW',
          'IY',  'EY',  'AA',  'ER',   'AW', 'AO',  'AY',  'OW',  'OY',
          'UH',  'IH',  'EH',  'AE',  'AH',  'UH',} | SLAX

## licit medial onsets

O2 = {('P', 'R'), ('T', 'R'), ('K', 'R'), ('B', 'R'), ('D', 'R'),
      ('G', 'R'), ('F', 'R'), ('TH', 'R'),
      ('P', 'L'), ('K', 'L'), ('B', 'L'), ('G', 'L'),
      ('F', 'L'), ('S', 'L'),
      ('K', 'W'), ('G', 'W'), ('S', 'W'),
      ('S', 'P'), ('S', 'T'), ('S', 'K'),
      ('HH', 'Y'), # "clerihew"
      ('R', 'W'),}
O3 = {('S', 'T', 'R'), ('S', 'K', 'L'), ('T', 'R', 'W')} # "octroi"

# This does not represent anything like a complete list of onsets, but
# merely those that need to be maximized in medial position.

def syllabify(pron, alaska_rule=True):
    """
    Syllabifies a CMU dictionary (ARPABET) word string

    # Alaska rule:
    >>> pprint(syllabify('AH0 L AE1 S K AH0'.split())) # Alaska
    '-AH0-.L-AE1-S.K-AH0-'
    >>> pprint(syllabify('AH0 L AE1 S K AH0'.split(), 0)) # Alaska
    '-AH0-.L-AE1-.S K-AH0-'

    # huge medial onsets:
    >>> pprint(syllabify('M IH1 N S T R AH0 L'.split())) # minstrel
    'M-IH1-N.S T R-AH0-L'
    >>> pprint(syllabify('AA1  K T R W AA0 R'.split())) # octroi
    '-AA1-K.T R W-AA0-R'

    # destressing
    >>> pprint(destress(syllabify('M IH1 L AH0 T EH2 R IY0'.split())))
    'M-IH-.L-AH-.T-EH-.R-IY-'

    # normal treatment of 'j':
    >>> pprint(syllabify('M EH1 N Y UW0'.split())) # menu
    'M-EH1-N.Y-UW0-'
    >>> pprint(syllabify('S P AE1 N Y AH0 L'.split())) # spaniel
    'S P-AE1-N.Y-AH0-L'
    >>> pprint(syllabify('K AE1 N Y AH0 N'.split())) # canyon
    'K-AE1-N.Y-AH0-N'
    >>> pprint(syllabify('M IH0 N Y UW2 EH1 T'.split())) # minuet
    'M-IH0-N.Y-UW2-.-EH1-T'
    >>> pprint(syllabify('JH UW1 N Y ER0'.split())) # junior
    'JH-UW1-N.Y-ER0-'
    >>> pprint(syllabify('K L EH R IH HH Y UW'.split())) # clerihew
    'K L-EH-.R-IH-.HH Y-UW-'

    # nuclear treatment of 'j'
    >>> pprint(syllabify('R EH1 S K Y UW0'.split())) # rescue
    'R-EH1-S.K-Y UW0-'
    >>> pprint(syllabify('T R IH1 B Y UW0 T'.split())) # tribute
    'T R-IH1-B.Y-UW0-T'
    >>> pprint(syllabify('N EH1 B Y AH0 L AH0'.split())) # nebula
    'N-EH1-B.Y-AH0-.L-AH0-'
    >>> pprint(syllabify('S P AE1 CH UH0 L AH0'.split())) # spatula
    'S P-AE1-.CH-UH0-.L-AH0-'
    >>> pprint(syllabify('AH0 K Y UW1 M AH0 N'.split())) # acumen
    '-AH0-K.Y-UW1-.M-AH0-N'
    >>> pprint(syllabify('S AH1 K Y AH0 L IH0 N T'.split())) # succulent
    'S-AH1-K.Y-AH0-.L-IH0-N T'
    >>> pprint(syllabify('F AO1 R M Y AH0 L AH0'.split())) # formula
    'F-AO1 R-M.Y-AH0-.L-AH0-'
    >>> pprint(syllabify('V AE1 L Y UW0'.split())) # value
    'V-AE1-L.Y-UW0-'

    # everything else
    >>> pprint(syllabify('N AO0 S T AE1 L JH IH0 K'.split())) # nostalgic
    'N-AO0-.S T-AE1-L.JH-IH0-K'
    >>> pprint(syllabify('CH ER1 CH M AH0 N'.split())) # churchmen
    'CH-ER1-CH.M-AH0-N'
    >>> pprint(syllabify('K AA1 M P AH0 N S EY2 T'.split())) # compensate
    'K-AA1-M.P-AH0-N.S-EY2-T'
    >>> pprint(syllabify('IH0 N S EH1 N S'.split())) # inCENSE
    '-IH0-N.S-EH1-N S'
    >>> pprint(syllabify('IH1 N S EH2 N S'.split())) # INcense
    '-IH1-N.S-EH2-N S'
    >>> pprint(syllabify('AH0 S EH1 N D'.split())) # ascend
    '-AH0-.S-EH1-N D'
    >>> pprint(syllabify('R OW1 T EY2 T'.split())) # rotate
    'R-OW1-.T-EY2-T'
    >>> pprint(syllabify('AA1 R T AH0 S T'.split())) # artist
    '-AA1 R-.T-AH0-S T'
    >>> pprint(syllabify('AE1 K T ER0'.split())) # actor
    '-AE1-K.T-ER0-'
    >>> pprint(syllabify('P L AE1 S T ER0'.split())) # plaster
    'P L-AE1-S.T-ER0-'
    >>> pprint(syllabify('B AH1 T ER0'.split())) # butter
    'B-AH1-.T-ER0-'
    >>> pprint(syllabify('K AE1 M AH0 L'.split())) # camel
    'K-AE1-.M-AH0-L'
    >>> pprint(syllabify('AH1 P ER0'.split())) # upper
    '-AH1-.P-ER0-'
    >>> pprint(syllabify('B AH0 L UW1 N'.split())) # balloon
    'B-AH0-.L-UW1-N'
    >>> pprint(syllabify('P R OW0 K L EY1 M'.split())) # proclaim
    'P R-OW0-.K L-EY1-M'
    >>> pprint(syllabify('IH0 N S EY1 N'.split())) # insane
    '-IH0-N.S-EY1-N'
    >>> pprint(syllabify('IH0 K S K L UW1 D'.split())) # exclude
    '-IH0-K.S K L-UW1-D'
    """
    ## main pass
    mypron = list(pron)
    nuclei = []
    onsets = []
    i = -1
    for (j, seg) in enumerate(mypron):
        if seg in VOWELS:
            nuclei.append([seg])
            onsets.append(mypron[i + 1:j]) # actually interludes, r.n.
            i = j
    codas = [mypron[i + 1:]]
    ## resolve disputes and compute coda
    for i in range(1, len(onsets)):
        coda = []
        # boundary cases
        if len(onsets[i]) > 1 and onsets[i][0] == 'R':
            nuclei[i - 1].append(onsets[i].pop(0))
        if len(onsets[i]) > 2 and onsets[i][-1] == 'Y':
            nuclei[i].insert(0, onsets[i].pop())
        if len(onsets[i]) > 1 and alaska_rule and nuclei[i-1][-1] in SLAX \
                                              and onsets[i][0] == 'S':
            coda.append(onsets[i].pop(0))
        # onset maximization
        depth = 1
        if len(onsets[i]) > 1:
            if tuple(onsets[i][-2:]) in O2:
                depth = 3 if tuple(onsets[i][-3:]) in O3 else 2
        for j in range(len(onsets[i]) - depth):
            coda.append(onsets[i].pop(0))
        codas.insert(i - 1, coda)

    output = list(zip(onsets, nuclei, codas)) 
    return output


def syllabify_pprint(syllab):
    """
    Pretty  #print a syllabification
    """
    return '.'.join('-'.join(' '.join(p) for p in syl) for syl in syllab)


def destress(syllab):
    """
    Generate a syllabification with nuclear stress information removed
    """
    syls = []
    for (onset, nucleus, coda) in syllab:
        nuke = [p[:-1] if p[-1] in {'0', '1', '2'} else p for p in nucleus]
        syls.append((onset, nuke, coda))
    return syls



import unittest
from ipclassifier.token_processors.spelling import SpellingNormalizer
from ipclassifier.utils import DictsSingleton



class TestSpellingNormalizer(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.dicts = DictsSingleton()
        cls.uk_us_dict = cls.dicts.uk_us_dict
        cls.regularize_dicts = cls.dicts.regularize_dicts


    def test_apostrophe_tis(self):
        sn = SpellingNormalizer("'tis", self.uk_us_dict, self.regularize_dicts)
        self.assertEqual(sn.modernized_word, ["tis", True, 'initial'])

    def test_apostrophe_th(self):
        sn = SpellingNormalizer("th'", self.uk_us_dict, self.regularize_dicts)
        self.assertEqual(sn.modernized_word, ["the", True, 'final'])

    def test_apostrophe_s(self):
        sn = SpellingNormalizer("heav'n", self.uk_us_dict, self.regularize_dicts)
        self.assertEqual(sn.modernized_word, ["heaven", True, 'medial'])

    def test_whettinge(self):
        sn = SpellingNormalizer("whettinge", self.uk_us_dict, self.regularize_dicts)
        self.assertEqual(sn.modernized_word, ["whetting", False, ''])

    def test_battelyons(self):
        sn = SpellingNormalizer("battelyons", self.uk_us_dict, self.regularize_dicts)
        self.assertEqual(sn.modernized_word, ["battalions", False, ''])

    def test_asdfasdf(self):
        sn = SpellingNormalizer("asdfasdf", self.uk_us_dict, self.regularize_dicts)
        self.assertEqual(sn.modernized_word, [None, False, ''])

    def test_viewest(self):
        sn = SpellingNormalizer("viewest", self.uk_us_dict, self.regularize_dicts)
        self.assertEqual(sn.modernized_word, ["view", False, ''])   

    def test_renewest(self):
        sn = SpellingNormalizer("renewest", self.uk_us_dict, self.regularize_dicts)
        self.assertEqual(sn.modernized_word, ["renew", False, ''])    

    def test_colour(self):
        sn = SpellingNormalizer("colour", self.uk_us_dict, self.regularize_dicts)
        self.assertEqual(sn.modernized_word, ["color", False, ''])     


if __name__ == "__main__":
    unittest.main()


import itertools


class CombinationsBaseMixin:
    """
    Creates new possible lines of IP given multiple versions of a token
    """

    def __init__(self, lines, product_possibles, is_odd_postition, target_stress):
        self._lines = lines
        self._product_possibles = product_possibles
        self._is_odd_position = 1 if is_odd_postition else 0
        self._target_stress = target_stress


    def _get_word_combinations(self, word, zero_stress_idxs):
        possible_words = []
        for possible in itertools.product(self._product_possibles,repeat=len(zero_stress_idxs)):
            word_copy = [s for s in word]
            for i,p in enumerate(possible):
                for j,s in enumerate(word_copy):
                    if j != zero_stress_idxs[i]:
                        word_copy[j] = s
                    else:
                        word_copy[j] = p            
            possible_words.append(word_copy)
        return possible_words


    def _get_line_combinations(self):
        line_pos = 0
        for line in self._lines:
            temp_line = []
            for word in line:
                if len(word) < 2:
                    line_pos += len(word)
                    temp_line.append([word])
                    continue
                zero_stress_idxs = [i for i,syl in enumerate(word) if (i + line_pos) % 2 == self._is_odd_position and syl == self._target_stress]
                if zero_stress_idxs:
                    word_combinations = self._get_word_combinations(word, zero_stress_idxs)
                    temp_line.append(word_combinations)
                else:
                    temp_line.append([word])
                line_pos += len(word)
            line_pos = 0
            self._build_graph(temp_line)


import unittest
from ipclassifier.token_processors.compounds import Compounds
from ipclassifier.utils import DictsSingleton



class TestCompounds(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.dicts = DictsSingleton()
        cls.words = cls.dicts.words
        cls.lemmatizer = cls.dicts.lemmatizer
        cls.uk_us_dict = cls.dicts.uk_us_dict
        cls.regularize_dicts = cls.dicts.regularize_dicts


    def test_compound_not_found(self):
        c = Compounds("apple", self.words, self.lemmatizer, self.uk_us_dict, self.regularize_dicts)
        compound = c.find_compound_in_wordlist()
        self.assertIsNone(compound)

    def test_compound_moonfaced_intercepted(self):
        c = Compounds("moonfaced", self.words, self.lemmatizer, self.uk_us_dict, self.regularize_dicts)
        compound = c.find_compound_in_wordlist()
        self.assertIsNone(compound)

    def test_compound_lampshine(self):
        c = Compounds("lampshine", self.words, self.lemmatizer, self.uk_us_dict, self.regularize_dicts)
        compound = c.find_compound_in_wordlist()
        self.assertEqual(compound, ['lamp', 'shine'])  

    def test_compound_selfstarter(self):
        c = Compounds("self-starter", self.words, self.lemmatizer, self.uk_us_dict, self.regularize_dicts)
        compound = c.find_compound_in_wordlist()
        self.assertEqual(compound, ['self', 'starter'])  

    def test_compound_colourcoded(self):
        c = Compounds("colour-shine", self.words, self.lemmatizer, self.uk_us_dict, self.regularize_dicts)
        compound = c.find_compound_in_wordlist()
        self.assertEqual(compound, ['color', 'shine'])  


if __name__ == "__main__":
    unittest.main()


from copy import deepcopy



class GraphBase():
    """
    Basic implementation of graph traversal to get all possible paths
    """

    def _build_graph(self, temp_line):
        graph = {i + 1: x for i,x in enumerate(temp_line) }
        for k,v in graph.items():
            graph[k] = tuple((k + 1,word) for word in v)
        graph[0] = ((1,[]),)
        graph[k+1] = ((),)
        self._graph = graph


    def _get_all_paths(self, current, end):
        if current == end:
            self._all_paths.append(deepcopy(self._path))
            return
        else:
            for sister in self._graph[current]:
                self._path.append(sister)
                self._get_all_paths(sister[0], end)
                self._path.pop()




from copy import deepcopy
from .spelling import SpellingNormalizer
from .compounds import Compounds
from .spelling_syllabify import SpellingSyllabifier




class PhonemeFSM():
    """
    Transforms a word token into phonemeic representation
    EX: "moon" ->  [['M', 'UW1', 'N']]
    """

    def __init__(self, token, dicts):
        self.final_phoneme_repr = []
        self.normalized_spelling = ''

        self._initial_token = token
        self._dicts = dicts 
        self._cmudict = self._dicts.cmudict
        self._uk_us_dict = self._dicts.uk_us_dict
        self._regularize_dicts = self._dicts.regularize_dicts
        self._has_apostrophe = False
        self.apostrophe_position = ''
        self._is_compound = [False, False]
        self._left_compound = ''
        self._right_compound = ''
        self._count = 0
        self._stress_assigned_diy = False

        self._lookup()


    def _check_ed(self, phonemes):
        if self._initial_token.endswith("ed") and len(self._initial_token) > 4 and not self._is_compound[0]:
            antepenult_letter = self._initial_token[-3]
            if antepenult_letter not in "aeiou":
                phonemes_copy = deepcopy(phonemes[0])
                phonemes_copy.insert(-1,'EH0')
                phonemes.append(phonemes_copy)
        return phonemes


    def _lookup(self, called_by_normalize=False, compound_token=None):
        token = self.normalized_spelling if self.normalized_spelling else self._initial_token
        token = compound_token if compound_token else token
        try:
            phonemes = self._cmudict.get(token, None)
            if phonemes is None: raise KeyError()
            phonemes = self._check_ed(phonemes)
            self._handle_success(phonemes)
        except KeyError:
            if called_by_normalize and self._count < 1:
                self._count += 1
                self._compound()
            elif called_by_normalize and self._count > 2:
                self._handle_failure()
            else:
                self._count += 1
                self._normalize(token)

    
    def _normalize(self, token):
        spelling_normalized, has_apostrophe, apostrophe_position = SpellingNormalizer(token, self._uk_us_dict, self._regularize_dicts).modernized_word
        if has_apostrophe: self._has_apostrophe = True
        if spelling_normalized: self.normalized_spelling = spelling_normalized
        if apostrophe_position: self.apostrophe_position = apostrophe_position
        if spelling_normalized:
            self._lookup(called_by_normalize=True)
        else:
            if any(self._is_compound):
                self._lookup(called_by_normalize=True)
            else:
                self._compound()
        

    def _compound(self):
        token = self.normalized_spelling if self.normalized_spelling else self._initial_token
        compound = Compounds(token, self._dicts.words, self._dicts.lemmatizer, self._uk_us_dict, self._regularize_dicts).find_compound_in_wordlist()
        if compound:
            left, right = compound[0], compound[1]
            self._is_compound[0] = True
            self._lookup(compound_token=left)
            if self._stress_assigned_diy: return
            self._is_compound[1] = True
            self._lookup(compound_token=right)
        else:
            self._handle_failure()


    def _apostrophe(self, phonemes):
        if not self._has_apostrophe: return phonemes
        if self.normalized_spelling == "the": return phonemes + [['DH']]
        if self.normalized_spelling == "to": return phonemes + [['T']]
        pop_position = 0 if self.apostrophe_position == "initial" else -1
        phonemes_copy = deepcopy(phonemes)
        reduced = []
        for word in phonemes_copy:
            filtered = [i for i,v in enumerate(word) if v[-1].isdigit()]
            word.pop(filtered[pop_position])
            reduced.append(word)
        return phonemes + reduced


    def _diy_stress_assignment(self):
        self._is_compound = [False, False]
        self._left_compound = None
        self._right_compound = None
        spelling_syllabifier = SpellingSyllabifier(self._initial_token)
        tentative_phonemes = spelling_syllabifier.tentative_phonemes
        self._stress_assigned_diy = True
        self._handle_success(tentative_phonemes)


    def _check_stress_reduction(self, phonemes):
        """
        [['K', 'AA1', 'N', 'JH', 'ER0', 'IH0', 'NG']] ->
        [['K', 'AA1', 'N', 'JH', 'ER0', 'IH0', 'NG'], ['K', 'AA1', 'N', 'JH', 'IH0', 'NG']]
        """
        if self._stress_assigned_diy: return phonemes
        new_phonemes = []
        for word in phonemes:
            for i in range(len(word) - 1):
                if word[i][-1].isdigit() and word[i+1][-1].isdigit():
                    word_copy = deepcopy(word)
                    word_copy.pop(i + 1)
                    new_phonemes.append(word_copy)
                    break
            new_phonemes.append(word)
        return new_phonemes
            

    def _handle_success(self, phonemes):
        phonemes = self._check_stress_reduction(phonemes)
        if self._stress_assigned_diy: 
            self.final_phoneme_repr = []
            if self.final_phoneme_repr:
                return
            else:
                self.final_phoneme_repr = phonemes
            return

        if any(self._is_compound):
            if self._is_compound[0]:
                self._left_compound = phonemes
                self._is_compound[0] = False
            if self._is_compound[1]:
                self._right_compound = phonemes
                self._is_compound[1] = False
        if self._left_compound and self._right_compound:
            combined = [lt + rt for lt in self._left_compound for rt in self._right_compound]
            self.final_phoneme_repr = self._apostrophe(combined)
            return

        elif self._has_apostrophe:
            reduced_phonemes = self._apostrophe(phonemes)
            self.final_phoneme_repr = reduced_phonemes
            return
        
        self.final_phoneme_repr = phonemes
        return

        
    def _handle_failure(self):
        self._diy_stress_assignment()
        return




import csv
import os



class UKAmerican:
    """
    Creates a UK-US spelling dictionary
    """
    
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            muh_dict = cls.create_uk_us_dict()
            cls._instance.uk_us_dict = muh_dict
        return cls._instance


    @staticmethod
    def create_uk_us_dict():
        with open(os.path.join(os.path.dirname(__file__), "files/uk_american_text.txt")) as f:
            reader = csv.reader(f, delimiter="\t")
            next(reader)
            next(reader)
            return {pair[0]:pair[1] for pair in reader}



import unittest
from ipclassifier.token_processors.spelling_syllabify import SpellingSyllabifier



class TestSpellingSyllabifier(unittest.TestCase):

    def test_qu_count(self):
        ss = SpellingSyllabifier("quality")
        self.assertEqual(ss._syllable_count, 3)

    def test_ion_count(self):
        ss = SpellingSyllabifier("inspections")
        self.assertEqual(ss._syllable_count, 3)

    def test_ion_not_applied(self):
        ss = SpellingSyllabifier("sion")
        self.assertEqual(ss._syllable_count, 2)

    def test_ae_count(self):
        ss = SpellingSyllabifier("aeneid")
        self.assertEqual(ss._syllable_count, 3)

    def test_repeated_vowel_count(self):
        ss = SpellingSyllabifier("look")
        self.assertEqual(ss._syllable_count, 1)

    def test_ou_count(self):
        ss = SpellingSyllabifier("thought")
        self.assertEqual(ss._syllable_count, 1)

    def test_ey_count(self):
        ss = SpellingSyllabifier("linsey")
        self.assertEqual(ss._syllable_count, 2)

    def test_ie_count(self):
        ss = SpellingSyllabifier("pixie")
        self.assertEqual(ss._syllable_count, 2)

    def test_ea_count(self):
        ss = SpellingSyllabifier("treat")
        self.assertEqual(ss._syllable_count, 1)

    def test_consonantal_y_count(self):
        ss = SpellingSyllabifier("yea")
        self.assertEqual(ss._syllable_count, 1)

    def test_final_ed_count(self):
        ss = SpellingSyllabifier("harbingered")
        self.assertEqual(ss._syllable_count, 3)

    def test_dipthong_and_vowel_count(self):
        ss = SpellingSyllabifier("galilaean")
        self.assertEqual(ss._syllable_count, 4)

    def test_y_and_2_vowels_count(self):
        ss = SpellingSyllabifier("yeoman")
        self.assertEqual(ss._syllable_count, 2)

    def test_oldfashioned_est_count(self):
        ss = SpellingSyllabifier("renewest")
        self.assertEqual(ss._syllable_count, 2)

    def test_ey_and_repeated_vowel_count(self):
        ss = SpellingSyllabifier("linsey-woolsey")
        self.assertEqual(ss._syllable_count, 4)

    def test_ey_and_repeated_vowel_count(self):
        ss = SpellingSyllabifier("raiment")
        self.assertEqual(ss._syllable_count, 2)

    def test_frighted(self):
        ss = SpellingSyllabifier("frighted")
        self.assertEqual(ss._syllable_count, 1)
        self.assertEqual(len(ss.tentative_phonemes), 3)

    def test_melodise(self):
        ss = SpellingSyllabifier("melodise")
        self.assertEqual(ss._syllable_count, 3)

    def test_vaunting(self):
        ss = SpellingSyllabifier("vaunting")
        self.assertEqual(ss._syllable_count, 2)

    def test_latian(self):
        ss = SpellingSyllabifier("latian")
        self.assertEqual(ss._syllable_count, 2)
        self.assertEqual(len(ss.tentative_phonemes), 2)

    def test_insubmissive(self):
        ss = SpellingSyllabifier("insubmissive")
        self.assertEqual(ss._syllable_count, 4)

    def test_rouses(self):
        ss = SpellingSyllabifier("rouses")
        self.assertEqual(ss._syllable_count, 2)     

    def test_uncloyed(self):
        ss = SpellingSyllabifier("uncloyed")
        self.assertEqual(ss._syllable_count, 2)   
        self.assertEqual(len(ss.tentative_phonemes), 3)

    def test_gudgeons(self):
        ss = SpellingSyllabifier("gudgeons")
        self.assertIn(['AH1', 'AH0'], ss.tentative_phonemes)     

    def test_leveret(self):
        ss = SpellingSyllabifier("leveret")
        self.assertIn(['AH1', 'AH0'], ss.tentative_phonemes)   


if __name__ == "__main__":
    unittest.main()


from setuptools import setup, find_packages

with(open("README.md")) as f:
    long_desc = f.read()

setup(
    name='ipclassifier',
    version='1.0.4',    
    description='Classifies English iambic pentameter poetry by period',
    long_description=long_desc,
    long_description_content_type="text/markdown",
    keywords=["NLP", "iambic pentameter", "meter", "poetry", "English poetry", "scansion", "poetry classification"],
    url='https://github.com/Chad-Mowbray/iamb-classifier',
    author='Chad Mowbray',
    author_email='mail@mail.com',
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
                    "click==8.0.1",
                    "joblib==1.0.1",
                    "nltk==3.6.2",
                    "numpy==1.21.2",
                    "regex==2021.8.28",
                    "scikit-learn==0.24.2",
                    "scipy==1.7.1",
                    "sklearn==0.0",
                    "threadpoolctl==2.2.0",
                    "tqdm==4.62.2",
                    "Unidecode==1.2.0 ",               
                      ],
    classifiers=[
        'License :: OSI Approved :: MIT License',       
        'Programming Language :: Python :: 3',
    ],
    
)


from itertools import product
from .combinations_graph import CombinationsGraph



class IambicLine():
    """
    Takes a tokenized line of IP
    Checks if the line is a possible line of valid IP using the following process:
            0. No changes
            1. Demote compound-word stresses
            2. Demote monosyllable stresses
            3. Promote monosyllable stresses
            4. Promote polysyllabic zero stresses 
            5. Demote polysyllabic zero stresses
            6. Invalid line
    If rule 5 applies, then primary stress on a polysyllabic word had to be demoted
    Such words are contained in self._changed_words
    """

    BASE_PATTERN = (0,1,0,1,0,1,0,1,0,1)

    def __init__(self, tokens, DEV=False):
        self._tokens = tokens
        self._original_stress_patterns_per_token = []
        self._formatted_list_of_realized_stress_patterns = []
        self._unique_dict_of_realized_stress_patterns = {}
        self._is_valid_pattern = False
        self._current_state = 0
        self._rules_applied = []
        self._syllables_per_line = []
        self._changed_words = []

        if not DEV: self._main()


    def __str__(self):
        return f"{self._is_valid_pattern}, {len(self._rules_applied)}, {self._syllables_per_line}, {self._changed_words}"


    @property
    def line_facts(self):
        return {
            "rules_applied": len(self._rules_applied),
            "syllables_per_line": self._syllables_per_line,
            "changed_words": self._changed_words,
            "words_per_line": len(self._tokens),
        }

    
    def _get_transformed_words(self, comparisons):
        for comparison in comparisons:
            abstract,realized = comparison
            original = self._unique_dict_of_realized_stress_patterns[abstract]            
            lines = self._get_possible_stress_patterns_per_token(self._get_original_stress_patterns_per_token())
            realized = original
            iterations = len(realized[0])
            words = []
            for i in range(iterations):
                intermediate = []
                for line in lines:
                    intermediate.append(line[i])
                words.append(intermediate)

            def finder_real(word_variations, realized):
                if len(realized) <= 1: return False
                for word in word_variations:
                    if len(word) != len(realized): continue
                    for i in range(len(realized)):
                        if word[i] == 1:
                            if realized[i] == 1:
                                return False
                return True

            changed = [finder_real(words[i], realized[0][i]) for i in range(len(words)) ]
            res = [pair[1] for i,pair in enumerate(self._original_stress_patterns_per_token) if changed[i]]
            self._changed_words = res


    def _test_base_pattern(self):
        comparisons = []
        for potential,original in self._unique_dict_of_realized_stress_patterns.items():
            if len(potential) != len(self.BASE_PATTERN): continue
            comparison = [syl for i,syl in enumerate(potential) if i % 2 == 0 and syl in [0,2] or i % 2 == 1 and syl in [1,2] ]
            if len(comparison) == len(self.BASE_PATTERN):
                comparisons.append([potential, original])
        if comparisons: 
            if len(self._rules_applied) >= 5: self._get_transformed_words(comparisons)
            return True
        return False


    def _get_original_stress_patterns_per_token(self):
        return [(t.stress_patterns, t.token) for t in self._tokens]


    def _get_possible_stress_patterns_per_token(self, original_stress_patterns_per_token):
        """
        returns: List[Tuples(Lists)]
        """
        return list(product(*[pair[0] for pair in original_stress_patterns_per_token]))


    def _create_formatted_list_of_realized_stress_patterns(self, stress_patterns_per_token):
        """
        accepts: List[Tuples(Lists)]
        produces: 
            -adds on to self._formatted_list_of_realized_stress_patterns -- List[Dictionaries]
            -adds unique values to unique_dict_of_realized_stress_patterns -- Dict{Tuple: List[Tuples(Lists)]}
        """
        formatted_list_of_realized_stress_patterns = []
        for pattern in stress_patterns_per_token:
            f = tuple( syllable for word in pattern for syllable in word)
            line = {
                "formatted": f,
                "original": pattern
            }
            formatted_list_of_realized_stress_patterns.append(line) 
        self._formatted_list_of_realized_stress_patterns.extend(formatted_list_of_realized_stress_patterns)
        self._create_unique_dict_of_realized_stress_patterns()


    def _create_unique_dict_of_realized_stress_patterns(self):
        for entry in self._formatted_list_of_realized_stress_patterns:
            if self._unique_dict_of_realized_stress_patterns.get(entry["formatted"]):
                if entry["original"] not in self._unique_dict_of_realized_stress_patterns[entry["formatted"]]:
                    self._unique_dict_of_realized_stress_patterns[entry["formatted"]].append(entry["original"])
            else:
                self._unique_dict_of_realized_stress_patterns[entry["formatted"]] = [entry["original"]]


    def _get_syllables_per_line(self):
        for pattern in self._unique_dict_of_realized_stress_patterns:
            if len(pattern) not in self._syllables_per_line:
                self._syllables_per_line.append(len(pattern))


    def _is_valid_IP(self):
        self._get_syllables_per_line()
        if self._test_base_pattern():
            return True
        else:
            return self._fit_to_IP()


    def _fit_to_IP(self):
        """
        After the initial check:
            1. Demote compound-word stresses
            2. Demote monosyllable stresses
            3. Promote monosyllable stresses
            4. Promote polysyllabic zero stresses 
            5. Demote polysyllabic zero stresses
        """

        phases = {
            1: self._demote_compound_stress,
            2: self._demote_monosyllable_stress,
            3: self._promote_monosyllable_stresses,
            4: self._promote_polysyllabic_zero_stresses,
            5: self._demote_polysyllabic_primary_stresses,
        }
        self._rules_applied.append(phases[self._current_state].__name__)
        try:
            if phases[self._current_state]():
                return True
            else:
                return False
        except KeyError:
            self._current_state += 1
            self._rules_applied.append("FAILURE")
            return False


    def _promote_monosyllable_stresses(self):
        new_combinations = []
        monosyllable_nonprimary_count = 0
        line_i = 0
        for line in self._get_baseline_before_alteration():
            monosyllable_nonprimary_idxs = []
            for i, word in enumerate(line):
                if len(word) == 1 and word[0] in [0,2] and line_i % 2 == 1:
                    monosyllable_nonprimary_count += 1
                    monosyllable_nonprimary_idxs.append(i)
                line_i += len(word)
            line_i = 0
            if monosyllable_nonprimary_idxs:
                for possible in product([1,0],repeat=len(monosyllable_nonprimary_idxs)):
                    line_copy = [w for w in line]
                    for i,val in enumerate(possible):
                        line_copy[monosyllable_nonprimary_idxs[i]] = [val]
                    new_combinations.append(tuple(line_copy))
        return self._check_validity_and_continue(new_combinations)


    @staticmethod
    def _create_demoted_compound_variations(word):
        word_stress_variations = []
        for i,syllable in enumerate(word):
            if syllable == 1:
                word_copy = [*word]
                word_copy[i] = 2
                word_stress_variations.append(word_copy)
                continue
        return word_stress_variations


    def _demote_compound_stress(self):
        new_combinations = []
        primary_count = 0
        for line in self._get_baseline_before_alteration():
            for word in line:
                for syllable in word:
                    if syllable == 1:
                        primary_count += 1
                    if primary_count > 1:
                        word_stress_variations = self._create_demoted_compound_variations(word)
                        idx_word_in_line = line.index(word)
                        for word_stress_variation in word_stress_variations:
                            line_copy = [w for w in line]
                            line_copy[idx_word_in_line] = word_stress_variation
                            line_copy = tuple(line_copy)
                            new_combinations.append(line_copy)
                primary_count = 0
            new_combinations.append(line)
        return self._check_validity_and_continue(new_combinations)


    def _demote_monosyllable_stress(self):
        """
        Works on a List[Tuples(Lists)]
        Creates a new List[Tuples(Lists)]

        Only demotes possibles in weak position
        """
        new_combinations = []
        monosyllable_primary_count = 0
        line_i = 0
        for line in self._get_baseline_before_alteration():
            monosyllable_primary_idxs = []
            for i, word in enumerate(line):
                if len(word) == 1 and word[0] == 1 and line_i % 2 == 0:
                    monosyllable_primary_count += 1
                    monosyllable_primary_idxs.append(i)
                line_i += len(word)
            line_i = 0
            if monosyllable_primary_idxs:
                for possible in product([1,0],repeat=len(monosyllable_primary_idxs)):
                    line_copy = [w for w in line]
                    for i,val in enumerate(possible):
                        line_copy[monosyllable_primary_idxs[i]] = [val]
                    new_combinations.append(tuple(line_copy))
        return self._check_validity_and_continue(new_combinations)


    def _promote_polysyllabic_zero_stresses(self):
        """
        check if the word requires more than one stress
        can promote appropriate 0s to 2s
        """
        new_combinations = []
        for line in self._get_baseline_before_alteration():
            cg = CombinationsGraph([line], [0,2], True, 0)
            single_combinations = cg.new_combinations
            for single_combination in single_combinations:
                new_combinations.append(single_combination)

        return self._check_validity_and_continue(new_combinations)


    def _demote_polysyllabic_primary_stresses(self):
        new_combinations = []
        for line in self._get_baseline_before_alteration():
            cg = CombinationsGraph([line], [2, 1], False, 1)
            single_combinations = cg.new_combinations
            for single_combination in single_combinations:
                new_combinations.append(single_combination)

        return self._check_validity_and_continue(new_combinations)


    def _get_baseline_before_alteration(self):
        return (l for line in self._unique_dict_of_realized_stress_patterns.values() for l in line)


    def _check_validity_and_continue(self, new_combinations):
        self._create_formatted_list_of_realized_stress_patterns(new_combinations)
        self._current_state += 1
        self._is_valid_pattern = self._is_valid_IP()
        return self._is_valid_pattern


    def _main(self):
        self._tokens = [t for t in self._tokens if t.token not in ["th'", "t'"]]  # remove elided "the"
        self._original_stress_patterns_per_token = self._get_original_stress_patterns_per_token()
        possible_stress_patterns_per_token = self._get_possible_stress_patterns_per_token(self._original_stress_patterns_per_token)
        self._is_valid_pattern = self._check_validity_and_continue(possible_stress_patterns_per_token)
        



from re import search
from ipclassifier.syllabify import syllabify, syllabify_pprint
from .phoneme_fsm import PhonemeFSM



class Token():
    """
    Accepts a word
    Represents features of a token
    """

    def __init__(self, token, dicts):
        self.token = token
        self._dicts = dicts
        self.stress_patterns = []
        self.modified_token = ''
        self.phoneme_reprs = []
        self.syllabifications = []

        self._main()


    def __str__(self):
        return f"Token instance: {self.token}"
        

    def __call__(self, *args, **kwargs):
        return {
            "original_token": self.token,
            "modified_token": self.modified_token,
            "stress_patterns": self.stress_patterns,
            "phoneme_reprs": self.phoneme_reprs,
            "syllabifications": self.syllabifications
        }


    def _get_phonemes_from_dict(self):
        token = self.token
        fsm = PhonemeFSM(token, self._dicts)
        phoneme_reprs = fsm.final_phoneme_repr
        self.phoneme_reprs = phoneme_reprs
        self.modified_token = fsm.normalized_spelling
       

    def _get_syllabification(self):
        for stress_repr in self.phoneme_reprs:
            syllabified = syllabify_pprint(syllabify(stress_repr, 0)).split('.')
            self.syllabifications.append(syllabified)


    def _get_stress_patterns(self):
        stress_patterns = []
        for syllabification in self.syllabifications:
            pattern = [int(search(r'[12]', syl).group()) if search(r'[12]', syl) else 0 for syl in syllabification]
            stress_patterns.append(pattern)
        self.stress_patterns = stress_patterns
  

    def _main(self):
        self._get_phonemes_from_dict()
        self._get_syllabification()
        self._get_stress_patterns()



